/// @addtogroup homologicalAlgebra
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file QuotientGradedModule.h
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2006 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#if !defined(_QUOTIENTGRADEDMODULE_H_)
#define _QUOTIENTGRADEDMODULE_H_

#include "homologicalAlgebra/HomologySignature.h"
#include "auxil/CRef.h"

template<typename freeModuleType>
class QuotientGradedModule;

template<typename freeModuleType2>
void swap(QuotientGradedModule<freeModuleType2>& A_qgm1,QuotientGradedModule<freeModuleType2>& A_qgm2);

// ********** class QuotientGradedModule ********** //
/*
  This class provides a container to store homology (the results of homology computations),
  which consists of a graded module of quotient modules.
*/
template<typename freeModuleType>
class QuotientGradedModule{
public:
  typedef typename freeModuleType::MatrixType MatrixType;
  typedef typename freeModuleType::GeneratorType GeneratorType;

  QuotientGradedModule():m_topDim(-1){};

  QuotientGradedModule(
    std::vector<QuotientModule<freeModuleType> >& A_quotientModuleVector
  ):
    m_topDim(A_quotientModuleVector.size()-1)
  {
    swap(A_quotientModuleVector,m_componentModule);
  };

  template<typename freeModuleType2>
  friend void swap(QuotientGradedModule<freeModuleType2>& A_qgm1,QuotientGradedModule<freeModuleType2>& A_qgm2);

  friend std::ostream& operator<<(std::ostream& out, const QuotientGradedModule& A_qgM){
    out << "--- Quotient Graded Module ---" << std::endl;
    for(int i=0;i<A_qgM.m_topDim+1;++i){
      out << "  Dimension " << i << std::endl;
      out << A_qgM.m_componentModule[i] << std::endl;
    }
    return out;
  }

  operator CRef<HomologySignature>(){
    CRef<HomologySignature> hsCR(new HomologySignature());
    for(unsigned int i=0;i<m_componentModule.size();++i) hsCR().push_back(FGAGrpSignature(m_componentModule[i]));
    return hsCR;
  }

  std::string descriptor() const{
    std::ostringstream out;
    for(int i=0;i<m_topDim+1;++i){
      out << "H_" << i << " = " << m_componentModule[i].descriptor() << std::endl;
    }
    return out.str();
  }

  const QuotientModule<freeModuleType>& component(int i){
    return m_componentModule[i];
  }

  int TopDim(){
    return m_topDim;
  }

private:
  int m_topDim;
  std::vector<QuotientModule<freeModuleType> > m_componentModule;
}; // ********** class QuotientGradedModule ********** //

template<typename freeModuleType2>
inline void swap(QuotientGradedModule<freeModuleType2>& A_qgm1,QuotientGradedModule<freeModuleType2>& A_qgm2){
  std::swap(A_qgm1.m_topDim,A_qgm2.m_topDim);
  std::swap(A_qgm1.m_componentModule,A_qgm2.m_componentModule);
}

#endif //_QUOTIENTGRADEDMODULE_H_
/// @}

