/// @addtogroup homologicalAlgebra
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file cubSetFunctors.hpp
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2006 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#if !defined(_CUBFUNCT_H_)
#define _CUBFUNCT_H_

#include <time.h>
#include "homologicalAlgebra/HomologySignature.h"
#include "homologicalAlgebra/homAlgFunctors.hpp"
#include "auxil/CRef.h"
#include "auxil/Functor.h"

// -------------------------------------------------------------------------------------- //
/*
  Class template CubSetFunctors is actually only a container for
  the definitions of functors (functions represented as objects)
  for various transformations of cubical sets.
  The template parameters P_CubSet and P_CubCelSet are assumed to be classes
  representing full cubical sets and cellural cubical sets (representable cubical sets)
  respectively. All the functors are static objects of the class associated
  to respective static functions defined in the class. The association is done
  via definition templates just after the definition of the class
*/
// -------------------------------------------------------------------------------------- //
int lastAcyclicSubsetSize=0;

template<typename P_CubSet, typename P_CubCelSet>
class CubSetFunctors{

  public:

  // Shorthands for various types of CubSet Functors
  typedef Functor<CRef<ZRFCComplex>,CRef<P_CubCelSet> > RFCComplex_BCubCelSet;
  typedef Functor<CRef<ZRFCComplex>,CRef<P_CubSet> > RFCComplex_BCubSet;
  typedef Functor<CRef<P_CubSet>,CRef<P_CubSet> > BCubSet_BCubSet;
  typedef Functor<CRef<P_CubCelSet>,CRef<P_CubSet> > BCubCelSet_BCubSet;
  typedef Functor<CRef<P_CubCelSet>,CRef<P_CubCelSet> > BCubCelSet_BCubCelSet;

  typedef typename P_CubSet::BoolBCI_Ptr BoolBCI_Ptr;
  typedef typename P_CubSet::BoolBI_Ptr BoolBI_Ptr;

  typedef typename P_CubSet::CubSetT_void_Ptr CubSetT_void_Ptr;
  typedef typename P_CubSet::CubSetT_CubSetT_Ptr CubSetT_CubSetT_Ptr;

  static CubSetT_void_Ptr shave;
  static CubSetT_CubSetT_Ptr acyclicSubspace;

  static CRef<P_CubCelSet> BCubCelSetFromBCubSet(CRef<P_CubSet> A_cubSetCR){
    Stopwatch swCollar;
    A_cubSetCR().addEmptyCollar();
    fcout << "Constructed collar in  " << swCollar << std::endl;
//  #if EMBEDDING_DIM==3
    if(A_cubSetCR().embDim()==3){
      Stopwatch swShaving;
      A_cubSetCR().shaveBI();
      fcout << "Shaving reduced the set to " << A_cubSetCR().cardinality() << " elementary cubes in " << swShaving << std::endl;
    }
//  #endif
    Stopwatch swConversion;
    CRef<P_CubCelSet> repSetCR( new P_CubCelSet(A_cubSetCR()) );
    fcout << "Changed cubSet to repSet in  " << swConversion << std::endl;
    return repSetCR;
  }

  static BCubCelSet_BCubSet BCubCelSet_From_BCubSet;

  // -------------------------------------------------------------------------------------- //

  static CRef<P_CubCelSet> CubCelSetReduce(CRef<P_CubCelSet> A_cubCelSetCR){
    Stopwatch swReduct;
    fcout << "Starting shaving via free face collapses." << std::endl;
    A_cubCelSetCR().shave();
    fcout << "Reduced in " << swReduct << " to " << A_cubCelSetCR().cardinality() << " cells.\n";

    Stopwatch swCoReduct;
    fcout << "Starting free coface collapses." << std::endl;
    A_cubCelSetCR().coReduceCompactSet();
    fcout << "Reduced in " << swCoReduct << " to " << A_cubCelSetCR().cardinality() << " cells.\n";

    return A_cubCelSetCR;

  }

  static BCubCelSet_BCubCelSet BCubCelSet_CR_BCubCelSet;

  // -------------------------------------------------------------------------------------- //

  static CRef<ZRFCComplex> ZRFCComplexFromCubSet(CRef<P_CubSet> A_repSetCR){

    Stopwatch sw;
    typename P_CubSet::PointCoordIterator it(A_repSetCR());
    typename P_CubSet::PointCoordIterator itEnd=A_repSetCR().end();
    std::vector<ElementaryCube> eCubes;
    int dim=A_repSetCR().embDim();
    for(it=A_repSetCR().begin();it<itEnd;++it){
      eCubes.push_back(ElementaryCube(it.coords(),dim));
    };
    CRef<ZRFCComplex> rfccCR( new ZRFCComplex(eCubes));
    fcout << "Chain complex construction of CubSet completed in " << sw  << std::endl;
    return rfccCR;

  }

  static RFCComplex_BCubSet RFCComplex_From_BCubSet;

  // -------------------------------------------------------------------------------------- //

  static CRef<ZRFCComplex> ZRFCComplexFromCubCelSet(CRef<P_CubCelSet> A_cubCelSetCR){

    Stopwatch sw;
    typename P_CubCelSet::PointCoordIterator it(A_cubCelSetCR());
    typename P_CubCelSet::PointCoordIterator itEnd=A_cubCelSetCR().end();
    std::set<ElementaryCell> cells;
    int dim=A_cubCelSetCR().embDim();
    for(it=A_cubCelSetCR().begin();it<itEnd;++it){
      cells.insert(ElementaryCell(it.coords(),dim));
    };
    CRef<ZRFCComplex> rfccCR( new ZRFCComplex(cells));
    fcout << "Chain complex construction of CubCelSet completed in " << sw  << std::endl;
    return rfccCR;

  }

  static RFCComplex_BCubCelSet RFCComplex_From_BCubCelSet;

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAlgebraicReductionsRandom(CRef<P_CubSet> A_setCR){
    return (
      homSign_ArRandom_ZRFCComplex *
      RFCComplex_From_BCubSet
    )(A_setCR);
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAlgebraicReductionsSorted(CRef<P_CubSet> A_setCR){
    return (
      homSign_ArSorted_ZRFCComplex *
      RFCComplex_From_BCubSet
    )(A_setCR);
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAlgebraicReductionsLocallySorted(CRef<P_CubSet> A_setCR){
    return (
      homSign_ArLocallySorted_ZRFCComplex *
      RFCComplex_From_BCubSet
    )(A_setCR);
  }

  // -------------------------------------------------------------------------------------- //
  // For general use:
  static CRef<HomologySignature> HomologyViaAlgebraicReductions(CRef<P_CubSet> A_setCR){
    return (
      homSign_ArRandom_ZRFCComplex *
      RFCComplex_From_BCubSet
    )(A_setCR);
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAcyclicSubspace(CRef<P_CubSet> A_cubSetCR){

    Stopwatch swCollar;
    A_cubSetCR().addEmptyCollar();

    Stopwatch swShaving;
    if(shave) (A_cubSetCR().*shave)();
    fcout << "Shaving reduced the set to " << A_cubSetCR().cardinality() << " elementary cubes in " << swShaving << std::endl;

    P_CubSet acyclicSubset(A_cubSetCR());
    Stopwatch swAcSubsp;
    // Now find the acyclic subset. Note: the acyclic subset is removed from the original set!
    int zeroBettiNumber=(A_cubSetCR().*acyclicSubspace)(acyclicSubset);
    lastAcyclicSubsetSize=acyclicSubset.cardinality();
    fcout << "Acyclic subsets of " << zeroBettiNumber << " components with " << lastAcyclicSubsetSize << " elementary cubes constructed in " << swAcSubsp << std::endl;

    // We got the zeroth betti numbers in zeroBettiNumber and now we store them in HomologySignature
    CRef<HomologySignature> homSignZeroCR(new HomologySignature());
    std::vector<int> t;
    homSignZeroCR().push_back(FGAGrpSignature(zeroBettiNumber,t));

    Stopwatch swExcision;
    { // to save memory, keep this local
      // Construct the neighborhood of set-acyclicSubset
      P_CubSet wrapped(A_cubSetCR());
      wrapped.wrap();
      // Important are only these cubes in the acyclic subset, which intersect the difference
      acyclicSubset*=wrapped;
      // Now add back the trimmed acyclic subset to the set (it was removed when constructing the acyclic subset)
      A_cubSetCR()+=acyclicSubset;
    }
    fcout << "Excision completed in " << swExcision << std::endl;

    Stopwatch swConversion;
    CRef<P_CubCelSet> bmpRepSetCR(new P_CubCelSet(A_cubSetCR()));
    { // to save memory, we keep this local
      P_CubCelSet repAcyclicSubset(acyclicSubset);
      bmpRepSetCR()-=repAcyclicSubset;
    }
    fcout << "Conversion to repset completed in " << swConversion << std::endl;

    // Finally we call homSignViaAR and add the resulting hom signature to the zeroth betti numbers
    CRef<HomologySignature> homSignCR=(homSign_AR_ZRFCComplex*RFCComplex_From_BCubCelSet)(bmpRepSetCR);
    homSignCR()+=homSignZeroCR();

    return homSignCR;
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAcyclicSubspaceHOM(CRef<P_CubSet> A_cubSetCR){
    P_CubSet::neighbAcyclicBCI=&P_CubSet::neighbAcyclicHOM;
    shave=0;
    acyclicSubspace=&P_CubSet::acyclicSubspaceBCI;
    return HomologyViaAcyclicSubspace(A_cubSetCR);
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAcyclicSubspaceHOMShaved(CRef<P_CubSet> A_cubSetCR){
    P_CubSet::neighbAcyclicBCI=&P_CubSet::neighbAcyclicHOM;
    shave=&P_CubSet::shaveBCI;
    acyclicSubspace=&P_CubSet::acyclicSubspaceBCI;
    return HomologyViaAcyclicSubspace(A_cubSetCR);
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAcyclicSubspaceSI(CRef<P_CubSet> A_cubSetCR){
    P_CubSet::neighbAcyclicBCI=&P_CubSet::neighbAcyclicSI;
    shave=0;
    acyclicSubspace=&P_CubSet::acyclicSubspaceBCI;
    return HomologyViaAcyclicSubspace(A_cubSetCR);
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAcyclicSubspaceSIShaved(CRef<P_CubSet> A_cubSetCR){
    P_CubSet::neighbAcyclicBCI=&P_CubSet::neighbAcyclicSI;
    shave=&P_CubSet::shaveBCI;
    acyclicSubspace=&P_CubSet::acyclicSubspaceBCI;
    return HomologyViaAcyclicSubspace(A_cubSetCR);
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAcyclicSubspaceSIR(CRef<P_CubSet> A_cubSetCR){
    P_CubSet::neighbAcyclicBCI=&P_CubSet::neighbAcyclicSIR;
    shave=0;
    acyclicSubspace=&P_CubSet::acyclicSubspaceBCI;
    return HomologyViaAcyclicSubspace(A_cubSetCR);
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAcyclicSubspaceSIRShaved(CRef<P_CubSet> A_cubSetCR){
    P_CubSet::neighbAcyclicBCI=&P_CubSet::neighbAcyclicSIR;
    shave=&P_CubSet::shaveBCI;
    acyclicSubspace=&P_CubSet::acyclicSubspaceBCI;
    return HomologyViaAcyclicSubspace(A_cubSetCR);
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAcyclicSubspaceLTD3(CRef<P_CubSet> A_cubSetCR){
    if(embeddingDim != 3) throw std::runtime_error("This executable has been incorrectly compiled for embeddingDim different from 3!");
    P_CubSet::neighbAcyclicBI=&P_CubSet::neighbAcyclicLT;
    shave=0;
    acyclicSubspace=&P_CubSet::acyclicSubspaceBI;
    return HomologyViaAcyclicSubspace(A_cubSetCR);
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaAcyclicSubspaceLTD3Shaved(CRef<P_CubSet> A_cubSetCR){
    if(embeddingDim != 3) throw std::runtime_error("This executable has been incorrectly compiled for embeddingDim different from 3!");
    P_CubSet::neighbAcyclicBI=&P_CubSet::neighbAcyclicLT;
    shave=&P_CubSet::shaveBI;
    acyclicSubspace=&P_CubSet::acyclicSubspaceBI;
    return HomologyViaAcyclicSubspace(A_cubSetCR);
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> HomologyViaRepSetReductions(CRef<P_CubSet> A_setCR){
    return (
      homSign_AR_ZRFCComplex *
      RFCComplex_From_BCubCelSet *
      BCubCelSet_CR_BCubCelSet *
      BCubCelSet_From_BCubSet
    )(A_setCR);
  }
};// class CubSetFunctors;

// Definition templates of static objects of CubSetFunctors class template

template<typename P_CubSet, typename P_CubCelSet>
typename CubSetFunctors<P_CubSet,P_CubCelSet>::BCubCelSet_BCubSet CubSetFunctors<P_CubSet,P_CubCelSet>::BCubCelSet_From_BCubSet(CubSetFunctors<P_CubSet,P_CubCelSet>::BCubCelSetFromBCubSet);

template<typename P_CubSet, typename P_CubCelSet>
typename CubSetFunctors<P_CubSet,P_CubCelSet>::BCubCelSet_BCubCelSet CubSetFunctors<P_CubSet,P_CubCelSet>::BCubCelSet_CR_BCubCelSet(CubSetFunctors<P_CubSet,P_CubCelSet>::CubCelSetReduce);

template<typename P_CubSet, typename P_CubCelSet>
typename CubSetFunctors<P_CubSet,P_CubCelSet>::RFCComplex_BCubSet CubSetFunctors<P_CubSet,P_CubCelSet>::RFCComplex_From_BCubSet(CubSetFunctors<P_CubSet,P_CubCelSet>::ZRFCComplexFromCubSet);

template<typename P_CubSet, typename P_CubCelSet>
typename CubSetFunctors<P_CubSet,P_CubCelSet>::RFCComplex_BCubCelSet CubSetFunctors<P_CubSet,P_CubCelSet>::RFCComplex_From_BCubCelSet(CubSetFunctors<P_CubSet,P_CubCelSet>::ZRFCComplexFromCubCelSet);

template<typename P_CubSet, typename P_CubCelSet>
typename CubSetFunctors<P_CubSet,P_CubCelSet>::CubSetT_void_Ptr CubSetFunctors<P_CubSet,P_CubCelSet>::shave(&P_CubSet::shaveBCI);

template<typename P_CubSet, typename P_CubCelSet>
typename CubSetFunctors<P_CubSet,P_CubCelSet>::CubSetT_CubSetT_Ptr CubSetFunctors<P_CubSet,P_CubCelSet>::acyclicSubspace(&P_CubSet::acyclicSubspaceBCI);

// -------------------------------------------------------------------------------------- //
#endif //_CUBFUNCT_H_

/// @}

