/// @addtogroup homologicalAlgebra
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file homAlgFunctors.hpp
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2006 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#if !defined(_ALGFUNCT_H_)
#define _ALGFUNCT_H_

#include <time.h>
#include "homologicalAlgebra/HomologySignature.h"
#include "auxil/CRef.h"
#include "auxil/Functor.h"

typedef FreeModule<int,Matrix<int,0,0> > ZFreeModule;
typedef ReducibleFreeChainComplex<ZFreeModule> ZRFCComplex;

// -------------------------------------------------------------------------------------- //

template<typename freeModuleType>
class HomAlgFunctors{

  public:

  typedef ReducibleFreeChainComplex<freeModuleType> RFCComplex;
  typedef Functor<CRef<HomologySignature>,CRef<RFCComplex> > HomSign_RFCComplex;

  static CRef<HomologySignature> homSignViaSmith(
   CRef<FreeChainComplex<freeModuleType> > A_fccCR
  ){
   typedef typename freeModuleType::MatrixType MatrixType;
   typedef typename freeModuleType::GeneratorType GeneratorType;
   int topDim=A_fccCR().topDim();
   if(topDim<0) return CRef<HomologySignature>(new HomologySignature());
   std::vector<MatrixType> cycles(topDim+1);
   std::vector<MatrixType> boundaries(topDim+1);
   // We begin with computing the cycles and boundaries from the boundary map
   // for dimension from topDim to 1
   for(int i=topDim;i>0;--i){
     MatrixType m(A_fccCR().boundaryHomomorphism(i).coefMatrix());
     kernelImage(
       m,
       cycles[i],
       boundaries[i-1]
     );
   }
   // The above procedure misses cycles in dimension zero and boundaries in dimension topDim
   // Boundaries in dimension topDim are zero anyway but we neeed to add cycles in dimension zero
   cycles[0]=MatrixType::Identity(A_fccCR().chainModule(0).numberOfGenerators());
   boundaries[topDim]=MatrixType(A_fccCR().chainModule(topDim).numberOfGenerators(),0);

   std::vector<QuotientModule<freeModuleType> > quotientModuleVector(topDim+1);
   for(int i=0;i<topDim+1;++i){
     QuotientModule<freeModuleType> qM(
       A_fccCR().chainModule(i),
       cycles[i],
       boundaries[i]
     );

     swap(quotientModuleVector[i],qM);
   }

   QuotientGradedModule<freeModuleType> homologyGradedModule(quotientModuleVector);
   return CRef<HomologySignature>(homologyGradedModule); //
  }

  static typename RFCComplex::ReducibleFreeChainComplex_void_Ptr reduce;

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> homSignViaAR(
   CRef<ReducibleFreeChainComplex<freeModuleType> > A_rfccCR
  ){

    typedef typename freeModuleType::MatrixType MatrixType;
    typedef typename freeModuleType::GeneratorType GeneratorType;

    CRef<HomologySignature> homSignCR(new HomologySignature());
    CRef<HomologySignature> homSignSmithCR(new HomologySignature());
    Stopwatch t;
    int pairsReduced=(A_rfccCR().*reduce)();
    fcout << "Algebraic reductions (KMS) of " << pairsReduced << " pairs in " << t << std::endl;

    // We first check if the reductions were total, i.e. if all generators are now in the kernel
    // of the boundary operator
    // If so, there is no need to apply Smith, which may take time
    // Otherwise we apply Smith.
    // However this is a temporary solution: in every case we should apply Smith to the complex remaining
    // after the removal of all kernel generators and then take the direct sum of the Smith outcome and the removed generators
    std::vector<std::vector<int> > simpleHomologyGenerators;

    bool fullyReduced=A_rfccCR().removeSimpleHomologyGenerators(simpleHomologyGenerators);

    // First prepare the part of homology signature based on simple homology generators
    for(unsigned int i=0;i<simpleHomologyGenerators.size();++i){
     int betti=simpleHomologyGenerators[i].size();
     // There are no torsion coefficients, so put only Betti numbers
     std::vector<int> empty;
     homSignCR().push_back( FGAGrpSignature(betti,empty ) );
    }

    // Now find the other generators, in particular torsion generators
    if(!fullyReduced){
     Stopwatch s;
     CRef<FreeChainComplex<ZFreeModule> > fccCR(new FreeChainComplex<ZFreeModule>(A_rfccCR()));
     homSignSmithCR=homSignViaSmith(fccCR);
     fcout << "Homology via Smith diagonalization in " << s  << std::endl;
    }
    homSignCR()+=homSignSmithCR();
    homSignCR().shrink();

    return homSignCR;
  }

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> homSignViaAR_Random(
   CRef<ReducibleFreeChainComplex<freeModuleType> > A_rfccCR
  ){
    reduce=&ReducibleFreeChainComplex<freeModuleType>::reduce;
    return homSignViaAR(A_rfccCR);
  }

  static HomSign_RFCComplex homSign_ArRandom_RFCComplex;

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> homSignViaAR_Sorted(
   CRef<ReducibleFreeChainComplex<freeModuleType> > A_rfccCR
  ){
    reduce=&ReducibleFreeChainComplex<freeModuleType>::reduceViaSort;
    return homSignViaAR(A_rfccCR);
  }

  static HomSign_RFCComplex homSign_ArSorted_RFCComplex;

  // -------------------------------------------------------------------------------------- //

  static CRef<HomologySignature> homSignViaAR_LocallySorted(
   CRef<ReducibleFreeChainComplex<freeModuleType> > A_rfccCR
  ){
    reduce=&ReducibleFreeChainComplex<freeModuleType>::reduceViaLocalSort;
    return homSignViaAR(A_rfccCR);
  }

  static HomSign_RFCComplex homSign_ArLocallySorted_RFCComplex;

  // -------------------------------------------------------------------------------------- //

};// class HomAlgFunctors

template<typename freeModuleType>
typename ReducibleFreeChainComplex<freeModuleType>::ReducibleFreeChainComplex_void_Ptr HomAlgFunctors<freeModuleType>::reduce(&ReducibleFreeChainComplex<freeModuleType>::reduce);

template<typename freeModuleType>
typename HomAlgFunctors<freeModuleType>::HomSign_RFCComplex HomAlgFunctors<freeModuleType>::homSign_ArRandom_RFCComplex(&HomAlgFunctors::homSignViaAR_Random);

template<typename freeModuleType>
typename HomAlgFunctors<freeModuleType>::HomSign_RFCComplex HomAlgFunctors<freeModuleType>::homSign_ArSorted_RFCComplex(&HomAlgFunctors::homSignViaAR_Sorted);

template<typename freeModuleType>
typename HomAlgFunctors<freeModuleType>::HomSign_RFCComplex HomAlgFunctors<freeModuleType>::homSign_ArLocallySorted_RFCComplex(&HomAlgFunctors::homSignViaAR_LocallySorted);

// Here various flavours for testing
HomAlgFunctors<ZFreeModule>::HomSign_RFCComplex& homSign_ArRandom_ZRFCComplex=HomAlgFunctors<ZFreeModule>::homSign_ArRandom_RFCComplex;
HomAlgFunctors<ZFreeModule>::HomSign_RFCComplex& homSign_ArSorted_ZRFCComplex=HomAlgFunctors<ZFreeModule>::homSign_ArSorted_RFCComplex;
HomAlgFunctors<ZFreeModule>::HomSign_RFCComplex& homSign_ArLocallySorted_ZRFCComplex=HomAlgFunctors<ZFreeModule>::homSign_ArLocallySorted_RFCComplex;
// and homSign_AR_ZRFCComplex is for the general use selected as the most likely best
HomAlgFunctors<ZFreeModule>::HomSign_RFCComplex& homSign_AR_ZRFCComplex=HomAlgFunctors<ZFreeModule>::homSign_ArRandom_RFCComplex;

// -------------------------------------------------------------------------------------- //
#endif //_ALGFUNCT_H_
/// @}

