/// @addtogroup homologicalAlgebra
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file readCubFile.hpp
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2006 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#if !defined(_READCUBFILE_H_)
#define _READCUBFILE_H_

#include <iostream>
#include <fstream>
#include <sstream>
#include <exception>
#include <new>
using namespace std;

#include "auxil/ofstreamcout.h"
extern ofstreamcout fcout;
#include "bitSet/EuclBitSet.hpp"
#include "bitSet/CubCellSetT.hpp"
#include "bitSet/EmbDimException.h"

template <typename BCubSet>
CRef<BCubSet> readCubFile(const char* fileName){
  std::fstream in;

  CRef<BCubSet> cubSetCR(new BCubSet(256));

  bool binaryFile=false;

  in.open(fileName,ios::in);
  if(!in.good()){
    std::ostringstream s;
    s << "Unable to open file " << fileName << " for reading " << std::endl;
    throw std::runtime_error(s.str());
  }

  char c=in.get();
  if(c=='B') binaryFile=true;
  in.close();

  if(binaryFile){
//    fcout << "Reading binary file\n";
    cubSetCR().readBmp(fileName);
  }else{
    Stopwatch t;
//    fcout << "Reading text file\n";
    in.open(fileName,ios::in);
    std::ios::pos_type startOfIn=in.tellg();
    try{
      in >> cubSetCR();
    }catch(DimException){ // This was not a full cubical set, so we try to read an ordinary cubical set and convert to a full cubical set
      const int dim=BCubSet::BaseClass::theDim;
      in.clear();
      in.seekg(startOfIn);
      RepSet<ElementaryCube> eCubeRepSet;
      in >> eCubeRepSet;
      if(dim!=eCubeRepSet.embeddingDimension()) throw EmbDimException("operator>>(istream&,EuclBitSetT<P_BitSet,dim>::importRepSetOfElementaryCube): incorrect embedding dimension");
      typedef typename BCubSet::BaseClass::BaseClass P_BitSet;
      CubCellSetT<EuclBitSetT<P_BitSet,dim> > cubCellSet(eCubeRepSet);
      BCubSet cubSet(cubCellSet);
      swap(cubSet,cubSetCR());
    }
    in.close();
    fcout << "Reading completed in " << t << std::endl;
  }
  int sizeCnt=cubSetCR().cardinality();
  if(sizeCnt==0){
    throw std::runtime_error("An empty set found on input ");
  }
  fcout << "Found " << sizeCnt << " full cubes with enclosing box: (";
  int prod=1;
  for(int i=0;i<embeddingDim;++i){
    int factor=cubSetCR().getUnpaddedWidth(i);
    prod*=factor;
    fcout << factor << " ";
  }
  int cf= sizeCnt*100/prod;
  fcout << ")\nCovering is: " << cf << "%\n";
  return cubSetCR;
}

// An adaptation for programs which want to store the bitmap directly
template <typename BCubSet>
CRef<BCubSet>  readCubFile(const char* fileName,const int*& dims, const char*& bitmap){
  CRef<BCubSet> cubSetCR=readCubFile<BCubSet>(fileName);
  dims=cubSetCR().Dimensions();
  bitmap=reinterpret_cast<const char*>(cubSetCR().getBitmap());
  return cubSetCR;
}

#endif //_READCUBFILE_H_


/// @}

