/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file staticContainer.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _StaticContainer_H
#define _StaticContainer_H

#include <stdexcept>
#include <cstdlib> // for size_t
#include <new>

namespace capd{
namespace vectalg{

template<typename Scalar, int capacity>
class Container
{
public:
   typedef Scalar ScalarType;
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   Container();
   explicit Container(int);
   Container(int,bool); // it does not insert zeros

   iterator begin();
   iterator end();
   const_iterator begin() const;
   const_iterator end() const;

   Container& operator=(const Container&);
   void resize(int newCapacity);

   ScalarType& operator[](int);
   const ScalarType& operator[](int) const;
   ScalarType& operator()(int);
   const ScalarType& operator()(int) const;
   friend void swap(Container<Scalar,capacity>& A_c1, Container<Scalar,capacity>& A_c2)
   {
     iterator b=A_c1.begin(), e=A_c1.end();
     iterator i = A_c2.begin();
     while(b!=e)
     {
        std::swap(*b,*i);
        ++b;
        ++i;
     }
   }

// memory allocation
   static void* operator new[](size_t sizeOfData, int);
   static void* operator new[](size_t sizeOfData);
   static void setDefaultDimension(int newDefaultDim){}
   static const int dimension() {return capacity;}

protected:
   ScalarType data[capacity];
};

// --------------- iterator selection --------------------- //

template<typename Scalar, int capacity>
inline typename Container<Scalar,capacity>::iterator Container<Scalar,capacity>::begin()
{
   return iterator(data);
}

template<typename Scalar, int capacity>
inline typename Container<Scalar,capacity>::iterator Container<Scalar,capacity>::end()
{
   return iterator(data+capacity);
}

template<typename Scalar, int capacity>
inline typename Container<Scalar,capacity>::const_iterator Container<Scalar,capacity>::begin() const
{
   return const_iterator(data);
}

template<typename Scalar, int capacity>
inline typename Container<Scalar,capacity>::const_iterator Container<Scalar,capacity>::end() const
{
   return const_iterator(data+capacity);
}

// ------------------------- indexing ------------------------ //

template<typename Scalar, int capacity>
inline Scalar& Container<Scalar,capacity>::operator[] (int i)
{
   return data[i];
}

template<typename Scalar, int capacity>
inline const Scalar& Container<Scalar,capacity>::operator[] (int i) const
{
   return data[i];
}

template<typename Scalar, int capacity>
inline Scalar& Container<Scalar,capacity>::operator() (int i)
{
   return data[i-1];
}

template<typename Scalar, int capacity>
inline const Scalar& Container<Scalar,capacity>::operator() (int i) const
{
   return data[i-1];
}

// ----------------- memory allocation ------------------------

template <typename Scalar, int capacity>
inline void* Container<Scalar,capacity>::operator new[](size_t sizeOfData)
{
   return ::operator new[](sizeOfData);
}

template <typename Scalar, int capacity>
inline void* Container<Scalar,capacity>::operator new[](size_t sizeOfData,int)
{
   return ::operator new[](sizeOfData);
}

// --------------- member definitions ----------------------------- //

template<typename Scalar, int capacity>
inline Container<Scalar,capacity>::Container()
{
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b=ScalarType(0);
      ++b;
   }
}

template<typename Scalar, int capacity>
inline Container<Scalar,capacity>::Container(int)
{
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b=ScalarType(0);
      ++b;
   }
}

template<typename Scalar, int capacity>
inline Container<Scalar,capacity>::Container(int,bool)
{}

template<typename Scalar, int capacity>
inline Container<Scalar,capacity>& Container<Scalar,capacity>::operator=(const Container& a_c)
{
   iterator b=begin(), e=end();
   const_iterator i = a_c.begin();
   while(b!=e)
   {
      *b = *i;
      ++b;
      ++i;
   }

   return *this;
}

template<typename Scalar, int capacity>
inline void Container<Scalar,capacity>::resize(int newCapacity)
{
   if(newCapacity!=capacity)
      throw std::range_error("Cannot change capacity of static container");
}

}} //namespace capd::vectalg

#endif

/// @}
