/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file vector.hpp
///
/// @author Marian Mrozek, Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#if !defined(_vector_hpp_)
#define _vector_hpp_

#include <cmath>
#include <stack>
#include <stdexcept>

#include "capd/minmax.h"
#include "capd/power.h"
#include "vectalg/vectorintf.h"

template<typename Scalar,int dim>
typename Vector<Scalar,dim>::ScalarType Vector<Scalar,dim>::euclNorm(void) const
{
   Scalar sum(0);
   const_iterator b=begin(), e=end();
   while(b!=e)
   {
      sum += static_cast<Scalar>(power(*b,2));
      ++b;
   }
   return Scalar(sqrt(nonnegativePart(sum)));
}

template<typename Scalar,int dim>
bool Vector<Scalar,dim>::normalize()
{
   Scalar n;
   if( ! (( n=euclNorm()) > 0.0) )
   {
      return false;
   }else{
      iterator b=begin(), e=end();
      while(b!=e)
      {
         *b /=n;
         ++b;
      }
      return true;
   }
}

//---------------------------constructors---------------------------//


template<typename Scalar,int dim>
Vector<Scalar,dim>::Vector(int A_dimension, ScalarType data[]) : ContainerType(A_dimension,true)
{
   for(int i=0; i<dimension(); ++i)
      (*this)[i] = data[i];
}

template<typename Scalar, int dim>
Vector<Scalar,dim>::Vector(Vector<double,dim>& A_vect) : ContainerType(A_vect.dimension(),true)
{
   iterator b=begin(), e=end();
   typename Vector<double,dim>::iterator i = A_vect.begin();
   while(b!=e)
   {
      *b = static_cast<Scalar>(*i);
      ++b;
      ++i;
   }
}


template<typename Scalar,int dim>
void Vector<Scalar,dim>::clear(void)
{
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = Scalar(0);
      ++b;
   }
}

//------------------ unary arithmetic operations ------------------//

template<typename Scalar, int dim>
Vector<Scalar,dim> abs(const Vector<Scalar,dim>& v)
{
   Vector<Scalar,dim> result(v.dimension(),true);
   typename Vector<Scalar,dim>::iterator b=result.begin(), e=result.end();
   typename Vector<Scalar,dim>::const_iterator i=v.begin();
   while(b!=e)
   {
      *b = capd::abs(*i);
      ++b;
      ++i;
   }
   return result;
}


template<typename Scalar, int dim>
Vector<Scalar,dim> operator-(const Vector<Scalar,dim>& v)
{
   Vector<Scalar,dim> result(v.dimension(),true);
   typename Vector<Scalar,dim>::iterator b=result.begin(), e=result.end();
   typename Vector<Scalar,dim>::const_iterator i=v.begin();
   while(b!=e)
   {
      *b = -(*i);
      ++b;
      ++i;
   }
   return result;
}

//-----------------------arithmetic operations----------------------//

template<typename Scalar, int dim>
Vector<Scalar,dim> operator+(const Vector<Scalar,dim>& v1,const Vector<Scalar,dim>& v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::range_error("operator+: Incompatible vector dimensions");
   Vector<Scalar,dim> result(v1.dimension(),true);
   typename Vector<Scalar,dim>::iterator b=result.begin(), e=result.end();
   typename Vector<Scalar,dim>::const_iterator b1=v1.begin(), b2=v2.begin();
   while(b!=e)
   {
      *b = (*b1) + (*b2);
      ++b;
      ++b1;
      ++b2;
   }
   return result;
}


template<typename Scalar, int dim>
Vector<Scalar,dim> operator-(const Vector<Scalar,dim>& v1,const Vector<Scalar,dim>& v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::range_error("operator-: Incompatible vector dimensions");
   Vector<Scalar,dim> result(v1.dimension(),true);
   typename Vector<Scalar,dim>::iterator b=result.begin(), e=result.end();
   typename Vector<Scalar,dim>::const_iterator b1=v1.begin(), b2=v2.begin();
   while(b!=e)
   {
      *b = (*b1) - (*b2);
      ++b;
      ++b1;
      ++b2;
   }
   return result;
}


template<typename Scalar,int dim>
Scalar operator*(const Vector<Scalar,dim>& v1,const Vector<Scalar,dim>& v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::range_error("operator*: Incompatible vector dimensions");
   Scalar result(0);
   typename Vector<Scalar,dim>::const_iterator b1=v1.begin(), b2=v2.begin(), e1=v1.end();
   while(b1!=e1)
   {
      result += (*b1) * (*b2);
      ++b1;
      ++b2;
   }
   return result;
}


template<typename Scalar, typename FactorType, int dim>
Vector<Scalar,dim> operator*(const Vector<Scalar,dim>& v,const FactorType& s)
{
   Vector<Scalar,dim> result(v.dimension(),true);
   typename Vector<Scalar,dim>::const_iterator b=v.begin(), e=v.end();
   typename Vector<Scalar,dim>::iterator i=result.begin();
   while(b!=e)
   {
      *i = (*b) * s;
      ++i;
      ++b;
   }
   return result;
}


template<typename Scalar, typename FactorType, int dim>
Vector<Scalar,dim> operator*(const FactorType& s,const Vector<Scalar,dim>& v)
{
   Vector<Scalar,dim> result(v.dimension(),true);
   typename Vector<Scalar,dim>::const_iterator b=v.begin(), e=v.end();
   typename Vector<Scalar,dim>::iterator i=result.begin();
   while(b!=e)
   {
      *i = s * (*b);
      ++i;
      ++b;
   }
   return result;
}


template<typename Scalar, typename FactorType, int dim>
Vector<Scalar,dim> operator/(const Vector<Scalar,dim>& v, const FactorType& s)
{
   Vector<Scalar,dim> result(v.dimension(),true);
   typename Vector<Scalar,dim>::const_iterator b=v.begin(), e=v.end();
   typename Vector<Scalar,dim>::iterator i=result.begin();
   while(b!=e)
   {
      *i = (*b) / s;
      ++i;
      ++b;
   }
   return result;
}


template<typename Scalar,int dim>
Vector<Scalar,dim> operator+(const Vector<Scalar,dim>& v,const Scalar& s)
{
   Vector<Scalar,dim> result(v.dimension(),true);
   typename Vector<Scalar,dim>::const_iterator b=v.begin(), e=v.end();
   typename Vector<Scalar,dim>::iterator i=result.begin();
   while(b!=e)
   {
      *i = (*b) + s;
      ++i;
      ++b;
   }
   return result;
}


template<typename Scalar,int dim>
Vector<Scalar,dim> operator-(const Vector<Scalar,dim> &v,const Scalar& s)
{
   Vector<Scalar,dim> result(v.dimension(),true);
   typename Vector<Scalar,dim>::const_iterator b=v.begin(), e=v.end();
   typename Vector<Scalar,dim>::iterator i=result.begin();
   while(b!=e)
   {
      *i = (*b) - s;
      ++i;
      ++b;
   }
   return result;
}

//----------------------assignments - vectors---------------------------//


template<typename Scalar,int dim>
Vector<Scalar,dim>& Vector<Scalar,dim>::operator+=(const Vector<Scalar,dim>& v)
{
   if(dimension()!=v.dimension())
      throw std::range_error("operator+=: Incompatible vector dimensions");
   iterator i=begin();
   const_iterator b=v.begin(), e=v.end();
   while(b!=e)
   {
      *i += *b;
      ++i;
      ++b;
   }
   return *this;
}


template<typename Scalar,int dim>
Vector<Scalar,dim>& Vector<Scalar,dim>::operator-=(const Vector& v)
{
   if(dimension()!=v.dimension())
      throw std::range_error("operator-=: Incompatible vector dimensions");
   iterator i=begin();
   const_iterator b=v.begin(), e=v.end();
   while(b!=e)
   {
      *i -= *b;
      ++i;
      ++b;
   }
   return *this;
}

//----------------------assignments - Scalars---------------------------//

template<typename Scalar,int dim>
Vector<Scalar,dim>& Vector<Scalar,dim>::operator=(const Scalar& s)
{
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = s;
      ++b;
   }
   return *this;
}


template<typename Scalar,int dim>
Vector<Scalar,dim>& Vector<Scalar,dim>::operator+=(const Scalar& s)
{
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b += s;
      ++b;
   }
   return *this;
}


template<typename Scalar,int dim>
Vector<Scalar,dim>& Vector<Scalar,dim>::operator-=(const Scalar& s)
{
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b -= s;
      ++b;
   }
   return *this;
}


template<typename Scalar,int dim>
Vector<Scalar,dim>& Vector<Scalar,dim>::operator*=(const Scalar& s)
{
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b *= s;
      ++b;
   }
   return *this;
}


template<typename Scalar,int dim>
Vector<Scalar,dim>& Vector<Scalar,dim>::operator/=(const Scalar& s)
{
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b /= s;
      ++b;
   }
   return *this;
}


template<typename Scalar,int dim>
void Vector<Scalar,dim>::sorting_permutation(intVectorType& perm)
{
   int i=0,j,k;

   if(dimension()!= perm.dimension())
      throw std::range_error("sorting_permutation: Incompatible vector dimensions");
   typename intVectorType::iterator b=perm.begin(), e=perm.end();
   while(b!=e)
   {
      *b=i;
      ++i;
      ++b;
   }

   for(i=0;i<dimension();i++)
      for(j=dimension()-1;j>i;j--)
      {
         if((*this)[perm[j]] > (*this)[perm[j-1]])
         {
            k=perm[j-1];
            perm[j-1]=perm[j];
            perm[j]=k;
         }
      }
}

//-------coord-wise inequalities - true if true for each coord---------//

template<typename Scalar,int dim>
bool operator< (const Vector<Scalar,dim>& v1, const Vector<Scalar,dim>& v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::range_error("operator<: Incompatible vector dimensions");
   typename Vector<Scalar,dim>::const_iterator b1=v1.begin(), b2=v2.begin(), e1=v1.end();
   while(b1!=e1)
   {
      if(!(*b1 < *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}


template<typename Scalar,int dim>
bool operator> (const Vector<Scalar,dim>& v1, const Vector<Scalar,dim>& v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::range_error("operator>: Incompatible vector dimensions");
   typename Vector<Scalar,dim>::const_iterator b1=v1.begin(), b2=v2.begin(), e1=v1.end();
   while(b1!=e1)
   {
      if(!(*b1 > *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}


template<typename Scalar,int dim>
bool operator<= (const Vector<Scalar,dim>& v1, const Vector<Scalar,dim>& v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::range_error("operator<=: Incompatible vector dimensions");
   typename Vector<Scalar,dim>::const_iterator b1=v1.begin(), b2=v2.begin(), e1=v1.end();
   while(b1!=e1)
   {
      if(!(*b1 <= *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}


template<typename Scalar,int dim>
bool operator>= (const Vector<Scalar,dim>& v1, const Vector<Scalar,dim>& v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::range_error("operator>=: Incompatible vector dimensions");
   typename Vector<Scalar,dim>::const_iterator b1=v1.begin(), b2=v2.begin(), e1=v1.end();
   while(b1!=e1)
   {
      if(!(*b1 >= *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}



template<typename Scalar,int dim>
bool operator== (const Vector<Scalar,dim>& v1, const Vector<Scalar,dim>& v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::range_error("operator==: Incompatible vector dimensions");
   typename Vector<Scalar,dim>::const_iterator b1=v1.begin(), b2=v2.begin(), e1=v1.end();
   while(b1!=e1)
   {
      if(!(*b1 == *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}


template<typename Scalar,int dim>
bool operator!= (const Vector<Scalar,dim>& v1, const Vector<Scalar,dim>& v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::range_error("operator<: Incompatible vector dimensions");
   typename Vector<Scalar,dim>::const_iterator b1=v1.begin(), b2=v2.begin(), e1=v1.end();
   while(b1!=e1)
   {
      if(!(*b1 != *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}


//----------------- input-output ----------------------------------//

template<typename Scalar, int dim>
std::ostream& operator<<(std::ostream& out, const Vector<Scalar,dim>& v)
{
   out << "{";
   if(v.dimension()>0){
     if(v[0]>=Scalar(0)) out << " ";
     out << v[0];
   }
   for(int i=1;i<v.dimension();i++)
   {
      out << ",";
      if(v[i]>=Scalar(0)) out << " ";
      out << v[i];
   }
   out << "}";
   return out;
}


template<typename Scalar, int dim>
std::istream& operator>> (std::istream& inp, Vector<Scalar,dim>& v)
{
   std::stack<Scalar> st;
   Scalar s;
   int ch;

   while('{'!=(ch=inp.get()) && ch!=EOF);
   if(ch!=EOF)
   {
      inp >> s;
      st.push(s);
      do{
         do{
            ch=inp.get();
         }while(ch==' ' || ch=='\t' || ch=='\n');
         if(ch==','){
            inp >> s;
            st.push(s);
         }
      }while(ch!='}' && ch!=EOF);
   }
   v.resize(st.size());
   int n=int(st.size());
   for(int i=0;i<n;++i)
   {
      v[n-i-1]=st.top();
      st.pop();
   }
   if(inp.eof()) throw std::ios_base::failure("EOF encountered when reading a vector");
   return inp;
}


#endif // _vector_hpp_

/// @}
